import json
import time
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
from typing import List
from sse_starlette.sse import EventSourceResponse

from openaoe.backend.model.dto.ClaudeDto import ClaudeChatReqDto, ClaudeMessage
from openaoe.backend.model.dto.ClaudeSlackDto import ClaudeSlackChatReqDto
from openaoe.backend.model.dto.ReturnBase import ReturnBase
from openaoe.backend.config.biz_config import claude_slack_cfg, claude_cfg
from openaoe.backend.util.choose_one import choose_one
from anthropic import Anthropic, HUMAN_PROMPT, AI_PROMPT
from openaoe.backend.config.constant import TYPE_BOT, TYPE_USER, TYPE_SYSTEM
from openaoe.backend.util.request_record_util import set_api_unhealthy
from openaoe.backend.config.constant import API_CLAUDE, VENDOR_CLAUDE
from openaoe.backend.util.request_info import *


def send_message(client, channel, text):
    try:
        return client.chat_postMessage(channel=channel, text=text)
    except SlackApiError as e:
        print(f"Error sending message: {e}")


def fetch_messages(client, channel, last_message_timestamp, bot_user_id):
    response = client.conversations_history(channel=channel, oldest=last_message_timestamp)
    return [msg['text'] for msg in response['messages'] if msg['user'] == bot_user_id]


def get_new_messages(client, channel, last_message_timestamp, bot_user_id):
    time.sleep(1)
    result_index = 0
    while True:
        messages = fetch_messages(client, channel, last_message_timestamp, bot_user_id)
        # 过滤Claude的Please note类型消息
        if messages and "*Please note:*" in messages[0]:
            result_index = 1

        if messages and len(messages) > result_index and not messages[result_index].endswith('Typing…_'):
            return messages[result_index]
        time.sleep(1)


def find_direct_message_channel(client, user_id):
    try:
        response = client.conversations_open(users=user_id)
        return response['channel']['id']
    except SlackApiError as e:
        print(f"Error opening DM channel: {e}")


def claude_chat_by_slack_svc(req_dto: ClaudeSlackChatReqDto):
    if not req_dto.token or not req_dto.appid:
        slack_infos = choose_one(claude_slack_cfg.get("slack_infos"))
        slack_user_token = slack_infos.split("+")[0]
        bot_user_id = slack_infos.split("+")[1]
    else:
        slack_user_token = req_dto.token
        bot_user_id = req_dto.appid
    client = WebClient(token=slack_user_token)
    dm_channel_id = find_direct_message_channel(client, bot_user_id)
    if not dm_channel_id:
        print("Could not find DM channel with the bot.")
        return

    last_message_timestamp = None
    prompt = req_dto.prompt
    response = send_message(client, dm_channel_id, prompt)
    if response:
        last_message_timestamp = response['ts']
    new_message = get_new_messages(client, dm_channel_id, last_message_timestamp, bot_user_id)

    return ReturnBase(
        data=new_message
    )


def claude_chat_anthropic(request: Request, req_dto: ClaudeChatReqDto):
    api_key = get_request_api_key(request, VENDOR_CLAUDE)
    prompt = gen_prompt(req_dto.prompt, req_dto.messages)
    if not prompt or len(prompt) == 0:
        return ReturnBase(
            msg="error",
            msgCode="-1",
            data="prompt or messages must be set"
        )

    try:
        anthropic = Anthropic(api_key=api_key)
        completion = anthropic.completions.create(
            model=req_dto.model,
            max_tokens_to_sample=req_dto.max_tokens,
            prompt=f"{HUMAN_PROMPT} {prompt}{AI_PROMPT}",
            temperature=req_dto.temperature
        )
        return ReturnBase(data=completion)
    except Exception as e:
        print(f"claude-chat failed: {e}")
        return ReturnBase(
            msg="error",
            msgCode="-1",
            data=str(e)
        )


def claude_chat_stream_svc(request, req_dto: ClaudeChatReqDto):
    api_key = get_request_api_key(request, VENDOR_CLAUDE)
    prompt = gen_prompt(req_dto.prompt, req_dto.messages)
    if not prompt or len(prompt) == 0:
        return ReturnBase(
            msg="error",
            msgCode="-1",
            data="prompt or messages must be set"
        )

    anthropic = Anthropic(api_key=api_key)

    async def stream():
        try:
            conn = anthropic.completions.create(
                prompt=prompt,
                max_tokens_to_sample=req_dto.max_tokens,
                model=req_dto.model,
                stream=True,
            )
            stop_flag = False
            while True:
                if await request.is_disconnected():
                    break

                for msg in conn:
                    dict_item = {
                        "msg": "",
                        "success": "true"
                    }

                    if msg.stop_reason:
                        dict_item["stop_reason"] = msg.stop_reason
                        stop_flag = True

                    if msg.completion:
                        dict_item["success"] = "true"
                        dict_item["msg"] = msg.completion

                    if dict_item:
                        yield json.dumps(dict_item, ensure_ascii=False)
                if stop_flag:
                    break
            return

        except Exception as e:
            # record api unhealthy
            set_api_unhealthy(API_CLAUDE)

            dict_item = {
                "success": "false",
                "msg": str(e)
            }
            yield json.dumps(dict_item)
            return

    return EventSourceResponse(stream())


def is_valid_model(model: str):
    valid_models = ["claude-1", "claude-1-100k", "claude-instant-1", "claude-instant-1-100k", "claude-1.3",
                    "claude-1.3-100k", "claude-1.0", "claude-instant-1.1-100k", "claude-instant-1.0"]
    return model in valid_models


def gen_prompt(prompt: str, messages: List[ClaudeMessage]):
    # 若messages为空，则使用prompt字段
    if not messages or len(messages) == 0:
        if not prompt or len(prompt) == 0:
            return ""
        return f"{HUMAN_PROMPT} {prompt}{AI_PROMPT}"

    ret = ""
    for item in messages:
        if item.role == TYPE_BOT:
            ret += f"{AI_PROMPT} {item.content} "
        elif item.role == TYPE_USER:
            ret += f"{HUMAN_PROMPT} {item.content} "
        elif item.role == TYPE_SYSTEM:
            ret = item.content + ret
    ret += f"{AI_PROMPT}"
    return ret


if __name__ == "__main__":
    # main()
    pass
